
import { unescape } from 'yizhi-html-escape'
import zlib from 'zlib'
import { genLyricTag, parseLyricTag } from './common'
import { IKaraokeWord, ILyric, ILyricLine, LyricType } from './declare'

const KUGOU_KEY = Buffer.from([64, 71, 97, 119, 94, 50, 116, 71, 81, 54, 49, 45, 206, 210, 110, 105])
const KRC_TAG = Buffer.from([0x6b, 0x72, 0x63, 0x31])

//加密、解密
function kugouConvert(buffer: Buffer) {
	for (let i = 0; i < buffer.byteLength; ++i) {
		buffer[i] ^= KUGOU_KEY[i % KUGOU_KEY.length]
	}
	return buffer
}

/**
 * 酷狗歌词解密，并返回krc文本
 * @param buffer 歌词内容
 */
export function decrypt(buffer: Buffer) {
	//检测开头4个字节
	if (buffer.compare(KRC_TAG, 0, KRC_TAG.length, 0, KRC_TAG.length)) return null
	try {
		//解密
		buffer = kugouConvert(buffer.subarray(4))
		//解压
		return zlib.inflateSync(buffer).toString()
	} catch (err) {
		return null
	}
}

/**
 * 加密krc歌词
 * @param krcText krc歌词文本
 */
export function encrypt(krcText: string) {
	//压缩
	let buffer = zlib.deflateSync(Buffer.from(krcText))
	//加密
	buffer = kugouConvert(buffer)
	//加上标记
	return Buffer.concat([KRC_TAG, buffer])
}

/**
 * 解析KRC歌词
 * @param krcText KRC歌词文本内容
 */
export function parse(krcText: string) {
	//结果
	const result: ILyric = { type: LyricType.KARA, content: [] }

	//逐行处理
	krcText.split(/\r?\n/).forEach(line => {
		line = line.trim()
		//歌词行
		const match = line.match(/^\[\s*(\d+)\s*,\s*(\d+)\s*\](.*)$/)
		if (match) {
			//目标信息
			const lrcLine: ILyricLine<IKaraokeWord[]> = {
				start: parseInt(match[1]),
				duration: parseInt(match[2]),
				content: [],
			}
			//一直往后所搜词语
			let rem = match[3]
			while (rem.length) {
				const match = rem.match(/^<(\d+),(\d+),(\d+)>([^<]+)/)
				if (match) {
					const word: IKaraokeWord = {
						start: parseInt(match[1]),
						duration: parseInt(match[2]),
						content: unescape(match[4]),
					}
					if (word.content) lrcLine.content.push(word)
					rem = rem.substring(match[0].length)
				}
				else break
			}
			//保存行
			if (lrcLine.content.length) result.content.push(lrcLine)
		}

		//标签
		else parseLyricTag(result, line)
	})

	return result
}

/**
 * 将歌词转换为KRC文本
 * @param lyric 歌词(只能是卡拉OK歌词)
 */
export function stringify(lyric: ILyric) {
	if (lyric.type != LyricType.KARA) throw new Error(`lrc cannot stringify to krc`)

	const buffer: string[] = [...genLyricTag(lyric)]

	lyric.content.forEach(line => {
		if (typeof line.content === 'string') return
		const dur = line.duration ?? (() => {
			const last = line.content[line.content.length - 1]
			if (!last) return 0
			return last.start + line.start + last.duration
		})()
		const text = line.content.map(word => `<${word.start},${word.duration},0>${word.content}`).join('')
		buffer.push(`[${line.start},${dur}]${text}`)
	})

	return buffer.join('\n')
}
